/*
 * Library interface to Linux x86 Performance-Monitoring Counters.
 *
 * Copyright (C) 1999  Mikael Pettersson
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <sys/ptrace.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include "libperfctr.h"

#define PAGE_SIZE	4096

static unsigned char cpu_type;
static unsigned pmc_himask;	/* 0xFF or 0xFFFF */
static int have_rdpmc = 0;	/* false on pre-MMX Intel P5 */
static int have_rdtsc = 0;	/* false on WinChip when PMCs are used */
static int dev_perfctr_fd = -1;

struct perfctr {
    int fd;
    volatile const struct perfctr_state *kstate;
};
static struct perfctr self = { -1, NULL };

static int check_init(void)
{
    if( dev_perfctr_fd < 0 ) {
	struct perfctr_info info;

	if( (dev_perfctr_fd = open("/dev/perfctr", O_RDONLY)) < 0 ) {
	    perror("open /dev/perfctr");
	    return -1;
	}
	if( ioctl(dev_perfctr_fd, PERFCTR_INFO, &info) < 0 )
	    return -1;
	cpu_type = info.cpu_type;
	switch( cpu_type ) {
	  case PERFCTR_CPU_CYRIX_MII:
	  case PERFCTR_CPU_AMD_K7:
	    pmc_himask = 0xFFFF;
	    break;
	  default:
	    pmc_himask = 0xFF;
	}
	have_rdpmc = (cpu_type != PERFCTR_CPU_INTEL_P5);
	switch( cpu_type ) {
	  case PERFCTR_CPU_WINCHIP_C6:
	  case PERFCTR_CPU_WINCHIP_2:
	    have_rdtsc = 0;
	    break;
	  default:
	    have_rdtsc = 1;
	}
    }
    return 1;
}

/*
 * Invokes the kernel driver directly.
 */

int perfctr_syscall(unsigned cmd, long arg)
{
    if( check_init() < 0 )
	return -1;
    return ioctl(dev_perfctr_fd, cmd, arg);
}

/*
 * Operations on explicitly attached perfctrs.
 */

static int
perfctr_attach_simple(int pid,
		      struct perfctr_control *control,
		      int writable,
		      struct perfctr *perfctr)
{
    int oerrno;
    unsigned cmd;

    cmd = writable ? PERFCTR_ATTACH_RDWR : PERFCTR_ATTACH_RDONLY;
    perfctr->fd = perfctr_syscall(cmd, pid);
    if( perfctr->fd < 0 )
	return -1;
    if( !control || !ioctl(perfctr->fd, PERFCTR_CONTROL, control) ) {
	perfctr->kstate = mmap(NULL, PAGE_SIZE, PROT_READ,
			       MAP_SHARED, perfctr->fd, 0);
	if( perfctr->kstate != MAP_FAILED )
	    return 0;
    }
    oerrno = errno;
    close(perfctr->fd);
    errno = oerrno;
    perfctr->fd = -1;
    perfctr->kstate = NULL;
    return -1;
}

static struct perfctr*
perfctr_attach_hairy(int pid, struct perfctr_control *control, int writable)
{
    int status, oerrno;
    struct perfctr *perfctr;

    perfctr = malloc(sizeof(struct perfctr));
    if( !perfctr )
	return perfctr;
    status = -1;
    if( pid == getpid() ) {
	status = perfctr_attach_simple(pid, control, writable, perfctr);
    } else if( ptrace(PTRACE_ATTACH, pid, 0, 0) == 0 ) {
	if( waitpid(pid, NULL, WUNTRACED) > 0 )
	    status = perfctr_attach_simple(pid, control, writable, perfctr);
	oerrno = errno;
	ptrace(PTRACE_DETACH, pid, 0, SIGCONT);
	errno = oerrno;
    }
    if( status < 0 ) {
	oerrno = errno;
	free(perfctr);
	errno = oerrno;
	perfctr = NULL;
    }
    return perfctr;
}

struct perfctr *perfctr_attach_rdonly(int pid)
{
    return perfctr_attach_hairy(pid, NULL, 0);
}

struct perfctr *perfctr_attach_rdwr(int pid, struct perfctr_control *control)
{
    return perfctr_attach_hairy(pid, control, 1);
}

int perfctr_read(const struct perfctr *perfctr, struct perfctr_state *state)
{
    do {
	*state = *perfctr->kstate;
    } while( state->start.tsc != perfctr->kstate->start.tsc );
    return 0;
}

int perfctr_control(const struct perfctr *perfctr,
		    struct perfctr_control *control)
{
    return ioctl(perfctr->fd, PERFCTR_CONTROL, control);
}

int perfctr_stop(const struct perfctr *perfctr)
{
    return ioctl(perfctr->fd, PERFCTR_STOP, NULL);
}

int perfctr_unlink(const struct perfctr *perfctr)
{
    return ioctl(perfctr->fd, PERFCTR_UNLINK, NULL);
}

void perfctr_close(struct perfctr *perfctr)
{
    if( perfctr->fd >= 0 ) {
	munmap((void*)perfctr->kstate, PAGE_SIZE);
	close(perfctr->fd);
	perfctr->kstate = NULL;
	perfctr->fd = -1;
    }
    if( perfctr != &self )
	free(perfctr);
}

/*
 * Operations on the process' own perfctrs.
 */

static int perfctr_attach_self(int writable)
{
    if( self.fd >= 0 ) {
	errno = EBUSY;
	return -1;
    }
    return perfctr_attach_simple(getpid(), NULL, writable, &self);
}

int perfctr_attach_rdonly_self(void)
{
    return perfctr_attach_self(0);
}

int perfctr_attach_rdwr_self(void)
{
    return perfctr_attach_self(1);
}

static __inline__ int check_self(void)
{
    return (self.fd < 0) ? perfctr_attach_rdwr_self() : 1;
}

union u64_conversion {	/* safe type-punning */
    unsigned int ui[2];
    unsigned long long ull;
};

static __inline__ unsigned long long rdtsc64(void)
{
    union u64_conversion u;
    __asm__ __volatile__("rdtsc" : "=a"(u.ui[0]), "=d"(u.ui[1]));
    return u.ull;
}

static __inline__ unsigned long long rdpmc64(unsigned int ctr)
{
    union u64_conversion u;
    __asm__ __volatile__("rdpmc; andl %3,%%edx"
			 : "=a"(u.ui[0]), "=d"(u.ui[1])
			 : "c"(ctr), "g"(pmc_himask));
    return u.ull;
}

int perfctr_read_self(struct perfctr_state *state)
{
    unsigned long long start_tsc;
    unsigned long long diff;
    struct perfctr_ctrs now;
    int i, nrctrs;

    if( check_self() < 0 )
	return -1;
    if( !have_rdpmc )		/* Intel pre-MMX P5 lossage */
	perfctr_syscall(PERFCTR_SAMPLE, 0);
    /* cannot use RDPMC unless the counters actually are running.. */
    nrctrs = self.kstate->status;
    if( !have_rdpmc || nrctrs <= 0 )
	return perfctr_read(&self, state);
    do {
	start_tsc = self.kstate->start.tsc;
	if( have_rdtsc )	/* WinChip lossage */
	    now.tsc = rdtsc64();
	for(i = 0; i < nrctrs; ++i)
	    now.pmc[i] = rdpmc64(i);
	if( have_rdtsc ) {	/* WinChip lossage */
	    diff = now.tsc - start_tsc;
	    state->sum.tsc = self.kstate->sum.tsc + diff;
	}
	for(i = 0; i < nrctrs; ++i) {
	    diff = now.pmc[i] - self.kstate->start.pmc[i];
	    state->sum.pmc[i] = self.kstate->sum.pmc[i] + diff;
	}
    } while( start_tsc != self.kstate->start.tsc );
    state->control = self.kstate->control;
    state->status = self.kstate->status;
    return 0;
}

int perfctr_control_self(struct perfctr_control *control)
{
    return perfctr_control(&self, control);
}

int perfctr_stop_self(void)
{
    return perfctr_stop(&self);
}

int perfctr_unlink_self(void)
{
    return perfctr_syscall(PERFCTR_UNLINK, 0);
}

void perfctr_close_self(void)
{
    return perfctr_close(&self);
}

/*
 * Miscellaneous operations.
 */

int perfctr_info(struct perfctr_info *info)
{
    return perfctr_syscall(PERFCTR_INFO, (long)info);
}

const char *perfctr_cpu_name(void)
{
    if( check_init() < 0 )
	return NULL;
    switch( cpu_type ) {
      case PERFCTR_CPU_INTEL_P5:
        return "Intel Pentium";
      case PERFCTR_CPU_INTEL_P5MMX:
        return "Intel Pentium MMX";
      case PERFCTR_CPU_INTEL_P6:
        return "Intel Pentium Pro";
      case PERFCTR_CPU_INTEL_PII:
        return "Intel Pentium II";
      case PERFCTR_CPU_INTEL_PIII:
        return "Intel Pentium III";
      case PERFCTR_CPU_CYRIX_MII:
        return "Cyrix 6x86MX or MII";
      case PERFCTR_CPU_WINCHIP_C6:
	return "WinChip C6";
      case PERFCTR_CPU_WINCHIP_2:
	return "WinChip 2";
      case PERFCTR_CPU_AMD_K7:
	return "AMD K7 Athlon";
      default:
        return "?";
    }
}

unsigned perfctr_evntsel_num_insns(void)
{
    /* This is terribly naive. Assumes only one event will be
     * selected, at CPL > 0.
     */
    if( check_init() < 0 )
	return (unsigned)-1;
    switch( cpu_type ) {
      case PERFCTR_CPU_INTEL_P5:
      case PERFCTR_CPU_INTEL_P5MMX:
      case PERFCTR_CPU_CYRIX_MII:
        /* event 0x16, count at CPL 3 */
        return 0x16 | (2 << 6);
      case PERFCTR_CPU_INTEL_P6:
      case PERFCTR_CPU_INTEL_PII:
      case PERFCTR_CPU_INTEL_PIII:
      case PERFCTR_CPU_AMD_K7:
        /* event 0xC0, count at CPL > 0, ENable */
        return 0xC0 | (1 << 16) | (1 << 22);
      case PERFCTR_CPU_WINCHIP_C6:
        return 0x02;
      case PERFCTR_CPU_WINCHIP_2:
        return 0x16;
      default:
        fprintf(stderr, "cpu_type %u not recognized\n", cpu_type);
        exit(1);
    }
}
